/*
 *  puzzlefw.cpp
 *
 *  C++ library for PuzzleFW firmware.
 *
 *  Joris van Rantwijk 2024
 */

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <filesystem>
#include <fstream>
#include <stdexcept>
#include <system_error>

#include "puzzlefw.hpp"


namespace fs = std::filesystem;


/** Path to platform devices. */
constexpr char SOC_DEVICES_DIR[] = "/sys/devices/soc0";


/**
 * Find PuzzleFW device in /sys filesystem.
 *
 * Return the device name.
 * Throw std::runtime_error if an error occurs.
 */
static std::string find_puzzlefw_device_name()
{
    // Find entry matching "*.puzzlefw" in /sys filesystem.
    for (const auto& entry : fs::directory_iterator(SOC_DEVICES_DIR)) {
        std::string fname = entry.path().filename().string();
        if (fname.compare(fname.size() - 9, 9, ".puzzlefw") == 0) {
            return fname;
        }
    }

    throw std::runtime_error(
        std::string("Device *.puzzlefw not found in ") + SOC_DEVICES_DIR);
}


/**
 * Find UIO device node for the PuzzleFW driver.
 *
 * Set "uio_path" to the full path to the UIO device node.
 * Set "dma_buf_size" to the size of the global DMA buffer.
 *
 * Throw std::runtime_error if an error occurs.
 */
static void find_puzzlefw_uio_device(const std::string& device_name,
                                     std::string& uio_path,
                                     size_t& dma_buf_size)
{
    // Construct path to device "uio" directory.
    fs::path uio_dir = fs::path(SOC_DEVICES_DIR) / device_name / "uio";

    // Find entry matching "uio*".
    fs::path uio_subdir;
    for (const auto& entry : fs::directory_iterator(uio_dir)) {
        fs::path p = entry.path();
        if (p.filename().string().compare(0, 3, "uio") == 0) {
            uio_subdir = p;
        }
    }

    if (uio_subdir.empty()) {
        throw std::runtime_error(
            std::string("No UIO device node found in ") + uio_dir.string());
    }

    // Construct full path to "/dev/uioN" node.
    fs::path uio_dev = fs::path("/dev") / uio_subdir.filename();
    uio_path = uio_dev.string();

    // Read DMA buffer size.
    std::ifstream map_size_if(uio_subdir / "maps" / "map1" / "size");
    unsigned long map_size;
    map_size_if >> std::hex >> map_size;
    dma_buf_size = map_size;
}


/* ********  class PuzzleFwDevice  ******** */

/* Constructor. */
puzzlefw::PuzzleFwDevice::PuzzleFwDevice()
  : m_uio_fd(-1)
  , m_regs(nullptr)
  , m_dma_buf(nullptr)
  , m_dma_buf_size(0)
{
    // Find platform device name of PuzzleFW firmware.
    m_device_name = find_puzzlefw_device_name();

    // Find UIO device node and DMA buffer size.
    find_puzzlefw_uio_device(m_device_name, m_uio_path, m_dma_buf_size);

    if (m_dma_buf_size % dma_alignment() != 0
            || m_dma_buf_size < MIN_DMA_BUF_SIZE) {
        throw std::runtime_error(std::string("Invalid DMA buffer size (")
                                 + std::to_string(m_dma_buf_size) + ")");
    }

    // Open UIO device node.
    // O_SYNC is necessary to map registers and DMA buffer as uncacheable.
    m_uio_fd = open(m_uio_path.c_str(), O_RDWR | O_SYNC);
    if (m_uio_fd < 0) {
        throw std::system_error(
            std::error_code(errno, std::system_category()),
            std::string("Can not open UIO device") + m_uio_path);
    }

    // Map FPGA user registers.
    // Offset 0 tells the UIO to map the first address range,
    // which corresponds to the FPGA registers.
    void *regs = mmap(nullptr,
                      puzzlefw::REGS_SIZE,
                      PROT_READ | PROT_WRITE,
                      MAP_SHARED,
                      m_uio_fd,
                      0);
    if (regs == ((void*)-1)) {
        close(m_uio_fd);
        throw std::system_error(
            std::error_code(errno, std::system_category()),
            std::string("Can not open mmap registers from ") + m_uio_path);
    }

    // Map DMA buffer.
    // Offset PAGE_SIZE tells the UIO to map the second address range,
    // which corresponds to the DMA buffer.
    int page_size = getpagesize();
    void *dma_buf = mmap(nullptr,
                         m_dma_buf_size,
                         PROT_READ | PROT_WRITE,
                         MAP_SHARED,
                         m_uio_fd,
                         1 * page_size);
    if (dma_buf == ((void*)-1)) {
        munmap(regs, puzzlefw::REGS_SIZE);
        close(m_uio_fd);
        throw std::system_error(
            std::error_code(errno, std::system_category()),
            std::string("Can not open mmap DMA buffer from ") + m_uio_path);
    }

    m_regs = (uint32_t*)regs;
    m_dma_buf = dma_buf;
}


/* Destructor. */
puzzlefw::PuzzleFwDevice::~PuzzleFwDevice()
{
    // Disable DMA engine.
    set_dma_enabled(false);

    // Unmap memory regions.
    munmap((void*)m_regs, puzzlefw::REGS_SIZE);
    munmap((void*)m_dma_buf, m_dma_buf_size);

    // Close UIO device.
    close(m_uio_fd);
}
